@@ -1,37 +1,18 @@
-"""
-This is a skeleton file that can serve as a starting point for a Python
-console script. To run this script uncomment the following lines in the
-``[options.entry_points]`` section in ``setup.cfg``::
-
-    console_scripts =
-         fibonacci = matbench_genmetrics.skeleton:run
-
-Then run ``pip install .`` (or ``pip install -e .`` for editable mode)
-which will install the command ``fibonacci`` inside your current environment.
-
-Besides console scripts, the header (i.e. until ``_logger``...) of this file can
-also be used as template for Python modules.
-
-Note:
-    This file can be renamed depending on your needs or safely removed if not needed.
-
-References:
-    - https://setuptools.pypa.io/en/latest/userguide/entry_point.html
-    - https://pip.pypa.io/en/stable/reference/pip_install
-"""
-
+"""Core functionality for matbench-genmetrics (generative materials benchmarking)"""
 import argparse
 import logging
 import sys
+from typing import List, Optional
 
 import numpy as np
 from mp_time_split.core import MPTimeSplit
-from pymatgen.analysis.structure_matcher import StructureMatcher
+from pymatgen.core.structure import Structure
 from scipy.stats import wasserstein_distance
 from tqdm import tqdm
 from tqdm.notebook import tqdm as ipython_tqdm
 
 from matbench_genmetrics import __version__
+from matbench_genmetrics.utils.match import ALLOWED_MATCH_TYPES, get_match_matrix
 
 # causes pytest to fail (tests not found, DLL load error)
 # from matbench_genmetrics.cdvae.metrics import RecEval, GenEval, OptEval
@@ -67,23 +48,35 @@ def fib(n):
     return a
 
 
-sm = StructureMatcher(stol=0.5, ltol=0.3, angle_tol=10.0)
-
-try:
-    import google.colab  # type: ignore # noqa: F401
-
-    IN_COLAB = True
-except ImportError:
-    IN_COLAB = False
+IN_COLAB = "google.colab" in sys.modules
 
 
 class GenMatcher(object):
-    def __init__(self, test_structures, gen_structures, verbose=True) -> None:
+    def __init__(
+        self,
+        test_structures,
+        gen_structures: Optional[List[Structure]] = None,
+        verbose=True,
+        match_type="cdvae_coverage",
+        **match_kwargs,
+    ) -> None:
         self.test_structures = test_structures
-        self.gen_structures = gen_structures
-        self.num_test = len(test_structures)
-        self.num_gen = len(gen_structures)
         self.verbose = verbose
+        assert (
+            match_type in ALLOWED_MATCH_TYPES
+        ), f"type must be one of {ALLOWED_MATCH_TYPES}"
+        self.match_type = match_type
+        self.match_kwargs = match_kwargs
+
+        if gen_structures is None:
+            self.gen_structures = test_structures
+            self.symmetric = True
+        else:
+            self.gen_structures = gen_structures
+            self.symmetric = False
+
+        self.num_test = len(self.test_structures)
+        self.num_gen = len(self.gen_structures)
 
         if self.verbose:
             # https://stackoverflow.com/a/58102605/13697228
@@ -102,10 +95,13 @@ class GenMatcher(object):
         if self._match_matrix is not None:
             return self._match_matrix
 
-        match_matrix = np.zeros((self.num_test, self.num_gen))
-        for i, ts in enumerate(self.tqdm(self.test_structures, **self.tqdm_kwargs)):
-            for j, gs in enumerate(self.gen_structures):
-                match_matrix[i, j] = sm.fit(ts, gs)
+        match_matrix = get_match_matrix(
+            self.test_structures,
+            self.gen_structures,
+            match_type=self.match_type,
+            symmetric=self.symmetric,
+            **self.match_kwargs,
+        )
 
         self._match_matrix = match_matrix
 
@@ -125,9 +121,12 @@ class GenMatcher(object):
 
     @property
     def duplicity_counts(self):
+        """Get number of duplicates within the match matrix ignoring self-comparison."""
         if self.num_test != self.num_gen:
             raise ValueError("Test and gen sets should be identical.")
-        # TODO: assert that test and gen sets are identical
+            # TODO: assert that test and gen sets are identical
+
+        # minus one is subtraction of the diagonal that got summed
         return np.clip(self.match_counts - 1, 0, None)
 
     @property
@@ -136,6 +135,12 @@ class GenMatcher(object):
 
     @property
     def duplicity_rate(self):
+        """Get number of duplicates divided by number of possible duplicate locations.
+
+        A set with 4 instances of the same structure will score lower than 2 repeat
+        instances each for 2 structures. In other words, the metric favors a larger
+        of unique structures, even if repeat structures exist.
+        """
         num_possible = self.num_test**2 - self.num_test
         return self.duplicity_count / num_possible
 
@@ -148,12 +153,16 @@ class GenMetrics(object):
         gen_structures,
         test_pred_structures=None,
         verbose=True,
+        match_type="cdvae_coverage",
+        **match_kwargs,
     ):
         self.train_structures = train_structures
         self.test_structures = test_structures
         self.gen_structures = gen_structures
         self.test_pred_structures = test_pred_structures
         self.verbose = verbose
+        self.match_type = match_type
+        self.match_kwargs = match_kwargs
         self._cdvae_metrics = None
         self._mpts_metrics = None
 
@@ -192,25 +201,41 @@ class GenMetrics(object):
     @property
     def coverage(self):
         """Match rate between test structures and generated structures."""
-        coverage = GenMatcher(
-            self.test_structures, self.gen_structures, verbose=self.verbose
-        ).match_rate
-        return coverage
+        self.coverage_matcher = GenMatcher(
+            self.test_structures,
+            self.gen_structures,
+            verbose=self.verbose,
+            match_type=self.match_type,
+            **self.match_kwargs,
+        )
+        return self.coverage_matcher.match_rate
 
     @property
     def novelty(self):
         """One minus match rate between train structures and generated structures."""
-        similarity = GenMatcher(
-            self.train_structures, self.gen_structures, verbose=self.verbose
-        ).match_rate
+        self.similarity_matcher = GenMatcher(
+            self.train_structures,
+            self.gen_structures,
+            verbose=self.verbose,
+            match_type=self.match_type,
+            **self.match_kwargs,
+        )
+        similarity = (
+            self.similarity_matcher.match_count / self.similarity_matcher.num_gen
+        )
         return 1.0 - similarity
 
     @property
     def uniqueness(self):
         """One minus duplicity rate within generated structures."""
-        commonality = GenMatcher(
-            self.gen_structures, self.gen_structures, verbose=self.verbose
-        ).duplicity_rate
+        self.commonality_matcher = GenMatcher(
+            self.gen_structures,
+            self.gen_structures,
+            verbose=self.verbose,
+            match_type=self.match_type,
+            **self.match_kwargs,
+        )
+        commonality = self.commonality_matcher.duplicity_rate
         return 1.0 - commonality
 
     @property
@@ -224,12 +249,23 @@ class GenMetrics(object):
         }
 
 
-class MPTSMetrics(GenMetrics):
-    def __init__(self, dummy=False, verbose=True):
+class MPTSMetrics(object):
+    def __init__(
+        self,
+        dummy=False,
+        verbose=True,
+        num_gen=None,
+        match_type="cdvae_coverage",
+        **match_kwargs,
+    ):
         self.dummy = dummy
         self.verbose = verbose
+        self.num_gen = num_gen
+        self.match_type = match_type
+        self.match_kwargs = match_kwargs
         self.mpt = MPTimeSplit(target="energy_above_hull")
         self.folds = self.mpt.folds
+        self.gms: List[Optional[GenMetrics]] = [None] * len(self.folds)
         self.recorded_metrics = {}
 
     def get_train_and_val_data(self, fold, include_val=False):
@@ -247,16 +283,61 @@ class MPTSMetrics(GenMetrics):
         return self.train_inputs
 
     def evaluate_and_record(self, fold, gen_structures, test_pred_structures=None):
-        GenMetrics.__init__(
-            self,
+        if self.num_gen is not None and self.num_gen != len(gen_structures):
+            raise ValueError(
+                f"Number of generated structures ({len(gen_structures)}) does not match expected number ({self.num_gen})."  # noqa: E501
+            )
+        self.gms[fold] = GenMetrics(
             self.train_inputs.tolist(),
             self.val_inputs.tolist(),
             gen_structures,
             test_pred_structures=test_pred_structures,
             verbose=self.verbose,
+            match_type=self.match_type,
+            **self.match_kwargs,
         )
 
-        self.recorded_metrics[fold] = self.metrics
+        self.recorded_metrics[fold] = self.gms[fold].metrics
+
+        # i.e. store the values for the current fold for testing purposes
+        for metric, value in self.recorded_metrics[fold].items():
+            setattr(self, metric, value)
+
+
+class MPTSMetrics10(MPTSMetrics):
+    def __init__(self, dummy=False, verbose=True):
+        MPTSMetrics.__init__(
+            self, dummy=dummy, verbose=verbose, num_gen=10, match_type="cdvae_coverage"
+        )
+
+
+class MPTSMetrics100(MPTSMetrics):
+    def __init__(self, dummy=False, verbose=True):
+        MPTSMetrics.__init__(
+            self, dummy=dummy, verbose=verbose, num_gen=100, match_type="cdvae_coverage"
+        )
+
+
+class MPTSMetrics1000(MPTSMetrics):
+    def __init__(self, dummy=False, verbose=True):
+        MPTSMetrics.__init__(
+            self,
+            dummy=dummy,
+            verbose=verbose,
+            num_gen=1000,
+            match_type="cdvae_coverage",
+        )
+
+
+class MPTSMetrics10000(MPTSMetrics):
+    def __init__(self, dummy=False, verbose=True):
+        MPTSMetrics.__init__(
+            self,
+            dummy=dummy,
+            verbose=verbose,
+            num_gen=10000,
+            match_type="cdvae_coverage",
+        )
 
 
 # def get_rms_dist(gen_structures, test_structures):
@@ -390,3 +471,20 @@ if __name__ == "__main__":
 #     match_rate = (match_rate + match_rate.T) / 2 - np.diag(np.diag(match_rate))
 
 #     return match_rate
+
+# from CBFV.composition import generate_features
+
+# self.test_formulas = [s.reduced_formula for s in test_structures]
+# self.gen_formulas = [s.reduced_formula for s in gen_structures]
+
+# self.test_cbfv, _, _, _ =
+# generate_features(pd.DataFrame(dict(formula=self.test_formulas, target=0.0)))
+# self.gen_cbfv, _, _, _ = generate_features(dict(formula=self.gen_formulas,
+# target=0.0))
+
+# try:
+#     import google.colab  # type: ignore # noqa: F401
+
+#     IN_COLAB = True
+# except ImportError:
+#     IN_COLAB = False
